/*
 * File : timestwo.c
 * Abstract:
 *       An example C-file S-function for multiplying an input by 2,
 *       y  = 2*u
 *
 * Real-Time Workshop note:
 *   This file can be used as is (noninlined) with the Real-Time Workshop
 *   C rapid prototyping targets, or it can be inlined using the Target 
 *   Language Compiler technology and used with any target. See 
 *     matlabroot/toolbox/simulink/blocks/tlc_c/timestwo.tlc   
 *   the TLC code to inline the S-function.
 *
 * See simulink/src/sfuntmpl_doc.c
 *
 * Copyright 1990-2009 The MathWorks, Inc.
 * $Revision: 1.1.6.2 $
 */


#define S_FUNCTION_NAME  fireCam
#define S_FUNCTION_LEVEL 2

/* number of s-function parameters */
#define N_PARS 1

#include "simstruc.h"

#include "fireCamLib.h"
#include <math.h>

#include <sys/time.h>
#include <time.h>


/*================*
 * Build checking *
 *================*/


#ifdef MATLAB_MEX_FILE 
#define MDL_CHECK_PARAMETERS 
/* Function: mdlCheckParameters =============================================
 * Abstract:
 *    Verify parameter settings.
 */
static void mdlCheckParameters(SimStruct *S) 
{
    /* par 1: steptime [double]
     */
    
    int i=0;
    
    for(i=0;i<N_PARS;i++)
    {
        boolean_T isOk = (mxIsNumeric(ssGetSFcnParam(S, i))  &&  
                          !mxIsLogical(ssGetSFcnParam(S, i)) &&
                          !mxIsComplex(ssGetSFcnParam(S, i)) &&  
                          !mxIsSparse(ssGetSFcnParam(S, i))  && 
                          !mxIsEmpty(ssGetSFcnParam(S, i))   &&
                          mxGetNumberOfElements(ssGetSFcnParam(S, i)) == 1);

        if(isOk){
            /* 
             * Translate the mxClassId of an mxArray to one of Simulink's
             * built-in data type ids. (see simulink.c)
             */
            DTypeId  dtype = ssGetDTypeIdFromMxArray(ssGetSFcnParam(S, i));
            if(i==0)
            {
                isOk = (dtype == SS_DOUBLE);
            }
            else
            {
                /*isOk = (dtype == SS_UINT32);*/
                isOk = (dtype == SS_DOUBLE);
            }
        }
        if(!isOk){   
            ssSetErrorStatus(S,"fireCam.c: Invalid s-function parameter.");
            return;
        }      
    }
}
#endif 

/* Function: mdlInitializeSizes ===============================================
 * Abstract:
 *   Setup sizes of the various vectors.
 */
static void mdlInitializeSizes(SimStruct *S)
{
    int_T i,ii;
    
    ssSetNumSFcnParams(S, N_PARS);
    if (ssGetNumSFcnParams(S) != ssGetSFcnParamsCount(S)) {
        return; /* Parameter mismatch will be reported by Simulink */
    }
    
    /* initialize input ports */
    if (!ssSetNumInputPorts(S, 4)) return;
    /* first input is shutter speed */
    ssSetInputPortWidth(S, 0, 1);
    ssSetInputPortDirectFeedThrough(S, 0, 1);
    ssSetInputPortDataType(S, 0, SS_DOUBLE);
    /* second input is white balance vector */
    ssSetInputPortWidth(S, 1, 2);
    ssSetInputPortDirectFeedThrough(S, 1, 1);
    ssSetInputPortDataType(S, 1, SS_UINT32);
    ssSetInputPortRequiredContiguous(S, 1, 1);
    /* third input is gain */
    ssSetInputPortWidth(S, 2, 1);
    ssSetInputPortDirectFeedThrough(S, 2, 1);
    ssSetInputPortDataType(S, 2, SS_UINT32);
    ssSetInputPortRequiredContiguous(S, 2, 1);
    /* fourth input is number of frames to save to disk */
    ssSetInputPortWidth(S, 3, 1);
    ssSetInputPortDirectFeedThrough(S, 3, 1);
    ssSetInputPortDataType(S, 3, SS_UINT32);
    ssSetInputPortRequiredContiguous(S, 3, 1);
    
    /* initialize output ports */
    if (!ssSetNumOutputPorts(S,4)) return;
    for(i=0; i<3; i++)
    {
        ssSetOutputPortDataType(S, i, SS_UINT8);
        ii = ssSetOutputPortMatrixDimensions(S, i, FRAME_WIDTH, FRAME_HEIGHT);
    }
    ssSetOutputPortWidth(S, 3, 1);
    ssSetOutputPortDataType(S, 3, SS_DOUBLE);
    
    /* Specify the number of sample times */
    ssSetNumSampleTimes(S, 1);

    /* set work vectors*/
    ssSetNumDWork(S,4);
    /* first DWork vector contains timestmp values and saved image count: 
     [0] is first image timestamp 
     [1] is previous image timestamp
     [2] is saved image count
     [3] is first real life timestamp
     */
    ssSetDWorkWidth(S, 0, 4);
    ssSetDWorkDataType(S, 0, SS_DOUBLE);
    /* second DWork vector contains previous shutter speed buffer */
    ssSetDWorkWidth(S, 1, 1);
    ssSetDWorkDataType(S, 1, SS_DOUBLE);
    /* third DWork vector contains white balance buffer */
    ssSetDWorkWidth(S, 2, 2);
    ssSetDWorkDataType(S, 2, SS_UINT32);
    /* fourth DWork vector contains gain buffer */
    ssSetDWorkWidth(S, 3, 1);
    ssSetDWorkDataType(S, 3, SS_UINT32);

    
    
   
}


/* Function: mdlInitializeSampleTimes =========================================
 * Abstract:
 *    Specifiy that we inherit our sample time from the driving block.
 */
static void mdlInitializeSampleTimes(SimStruct *S)
{
    ssSetSampleTime(S, 0, INHERITED_SAMPLE_TIME); 
    ssSetOffsetTime(S, 0, 0.0);
}



#define MDL_START  /* Change to #undef to remove function */
#if defined(MDL_START) 
  /* Function: mdlStart =======================================================
   * Abstract:
   *    This function is called once at start of model execution. If you
   *    have states that should be initialized once, this is the place
   *    to do it.
   */
  static void mdlStart(SimStruct *S)
  {
    int i;
    real_T* t = ssGetDWork(S, 0);
    
    /* stepsize */
    const real_T *pPtr0 = mxGetPr(ssGetSFcnParam(S,0));
    const real_T stepTime = *pPtr0;
    
    
    /* Camera prosilica EC750C is capable of a framerate 
     * between 2 fps and 60 fps for full resolution.
     * Ensure the steptime is allowed to comply with this range.*/
    if(stepTime < 0.0166 || stepTime > 0.5)
        ssSetErrorStatus(S,"\nfireCam.c ERROR: Invalid steptime specified, steptime must be fixed and obey 1/60 s < steptime < 1/2 s\n");
    
    /* initialize camera */
    #ifndef  MATLAB_MEX_FILE
    fireCamInit((int) 10/stepTime);
    printf("\nfireCam.c: Camera initialized\n");
    flushBuffer();
    #endif
         
    /* initilize work vector */
    for(i=1; i< ssGetDWorkWidth(S,0); i++)
        t[i] = 0.0;
    
  }
#endif /*  MDL_START */


#define MDL_UPDATE   /* Change to #undef to remove function */
#if defined(MDL_UPDATE)
/* Function: mdlUpdate =======================================================
 */
static void mdlUpdate(SimStruct *S, int_T tid) 
{
    /* buffer access */
    real_T* shutterSpeedBuffer = ssGetDWork(S, 1);
    InputRealPtrsType shutterSpeedNew = ssGetInputPortRealSignalPtrs(S,0);

    uint32_T* whiteBalanceBuffer = ssGetDWork(S, 2);
    uint32_T* whiteBalanceNew = (uint32_T*) ssGetInputPortSignal(S,1);

    uint32_T* gainBuffer = ssGetDWork(S, 3);
    uint32_T* gainNew = (uint32_T*) ssGetInputPortSignal(S,2);
    
    /* stepsize */
    const real_T *pPtr0 = mxGetPr(ssGetSFcnParam(S,0));
    const real_T stepTime = *pPtr0;
    
    /* change camera shutter speed setting if shutter speed input is changed */
    if(*shutterSpeedNew[0] != *shutterSpeedBuffer)
    {
        if(setShutterSpeedAbs(*shutterSpeedNew[0], stepTime))
        {
            *shutterSpeedBuffer = *shutterSpeedNew[0];
            printf("\nfireCam.c: Shutter speed set: %f [s]\n",*shutterSpeedNew[0]);
        }
        else
            printf("\nfireCam.c FAILURE: Could not set shutter speed: %f [s]\n",*shutterSpeedNew[0]);
    }
    
    /* change camera white balance settings if white balance input is changed */
    if(whiteBalanceNew[0] != whiteBalanceBuffer[0] || whiteBalanceNew[1] != whiteBalanceBuffer[1])
    {
        if(setWhiteBalance(whiteBalanceNew[0], whiteBalanceNew[1]))
        {
            whiteBalanceBuffer[0] = whiteBalanceNew[0];
            whiteBalanceBuffer[1] = whiteBalanceNew[1];
            printf("\nfireCam.c: White balance set: {blue red} = {%d %d}\n",whiteBalanceNew[0], whiteBalanceNew[1]);
        }
        else
            printf("\nfireCam.c FAILURE: Could not set white balance\n");
    }
    
    /* change camera gain setting if gain input is changed */
    if(*gainNew != *gainBuffer)
    {
        if(setGain((uint32_T)*gainNew))
        {
            *gainBuffer = *gainNew;
            printf("\nfireCam.c: gain set: %d\n",*gainNew);
        }
        else
            printf("\nfireCam.c FAILURE: Could not set gain\n");
    }
} 
#endif /*  MDL_UPDATE */


/* Function: mdlOutputs =======================================================
 */
static void mdlOutputs(SimStruct *S, int_T tid)
{
    /* general variables */
    int_T i, i_ts, fl_old;
    uint64_t timestamp=0;
    int_T frameLag;
    
    /* variables for frame saving */
    FILE* imagefile;
    char imgName[50];
    char dirName[30];
    char curTimeStr[30];
    struct timeval tv;
    time_t curTimeSec;

    /* Get number of frames to save */ 
    uint32_T* nFramesToSave = (uint32_T*) ssGetInputPortSignal(S,3);

    /* get pointer to work vector for timestamp storage
     [0] is first image timestamp 
     [1] is previous image timestamp
     [2] is saved image count
     [3] is first real life timestamp */
    real_T* ts = ssGetDWork(S, 0);
    
    /* initialize image pointers */
    uint8_T image[FRAME_WIDTH*FRAME_HEIGHT*3];
    uint8_T *R    = (uint8_T*) ssGetOutputPortSignal(S,0);
    uint8_T *G    = (uint8_T*) ssGetOutputPortSignal(S,1);
    uint8_T *B    = (uint8_T*) ssGetOutputPortSignal(S,2);
    
    /* image timestamp */
    real_T *img_ts    = (real_T*) ssGetOutputPortSignal(S,3);

    #ifndef  MATLAB_MEX_FILE
    /* get one frame from the image ringbuffer
     * - image is the image buffer
     * - timestamp is the represents the time when image is dumped in buffer [usec]
     * - frameLag is the number of frames lagging, should be 0 in case system is quick enough
     */
    frameLag = grabOneFrame(image, &timestamp);
    /* */
    
    
    
    /* check whether frame lag occurs */
    if(frameLag > 0) /* 2do: if result > 0 stop experiment because of missed frame */
        printf("\nfireCam.c: Frame lag: %d frame(s)\n",frameLag);
    /* */
    
    
    
    /* seperate image in seperate channels RGB and copy to outputs*/
    for (i=0; i<FRAME_WIDTH*FRAME_HEIGHT; i++)
    {
        R[i] = image[3*i];
        G[i] = image[3*i+1];
        B[i] = image[3*i+2];
    }
    /* */
    
    
    
    /* save current frame to file, if desired */
    if(nFramesToSave[0] > 0 && (uint32_T) ts[2] < nFramesToSave[0])
    {
        /* create directory on first step */
        if((uint32_T) ts[2] == 0)
        {
            /* get current time */
            gettimeofday(&tv, NULL); 
            curTimeSec=tv.tv_sec;
            strftime(curTimeStr,30,"%Y-%m-%d_%T",localtime(&curTimeSec));
            /* make directory */
            sprintf(dirName,"images_%s",curTimeStr);
            if(mkdir(dirName,0777) != 0)
            {
                printf("fireCam.c: Can't create create directory %s\n",dirName);
                ssSetErrorStatus(S,"fireCam: Can't create directory. Is executable running as superuser?");
                return;
            }
            /* save first timestamp [seconds] to working vector*/
            ts[3] = (real_T) curTimeSec;
        }
        
        /* generate time string from saved timestamp */
        curTimeSec=(time_t)ts[3];
        strftime(curTimeStr,30,"%Y-%m-%d_%T",localtime(&curTimeSec));
        
        /* save image */
        sprintf(imgName,"images_%s/image%d.ppm", curTimeStr, (uint32_T)ts[2]);
        imagefile=fopen(imgName, "wb");

        if(imagefile == NULL) 
        {
            printf( "fireCam: Can't create output file %s\n",imgName);
            ssSetErrorStatus(S,"fireCam: Can't create output file. Is executable running as superuser?");
            return;
        }
        else
        {
            fprintf(imagefile,"P6\n%u %u\n255\n", FRAME_WIDTH, FRAME_HEIGHT);
            fwrite(image, 1, FRAME_HEIGHT*FRAME_WIDTH*3, imagefile);
            fclose(imagefile);
        }
        
        /* update image counter */
        ts[2]++;
    }
    /* */
    
    
    
    /* store timestamps */
    if(ts[0] == 0)
        ts[0] = (real_T) timestamp;
    ts[1] = (real_T) timestamp;
    /* */
    
    
    
    /* set relative image timestamp as output */
    *img_ts = (real_T) (timestamp - ts[0]) / 1000000.0;
    /* */
    
    
    
    #else
    for (i=0; i<FRAME_WIDTH*FRAME_HEIGHT; i++)
    {
        *R++ = 1;
        *G++ = 2;
        *B++ = 3;
    }
    #endif
}


/* Function: mdlTerminate =====================================================
 * Abstract:
 *    No termination needed, but we are required to have this routine.
 */
static void mdlTerminate(SimStruct *S)
{
    /* close camera */
    #ifndef  MATLAB_MEX_FILE
    fireCamTerm();
    #endif
    
}



#ifdef  MATLAB_MEX_FILE    /* Is this file being compiled as a MEX-file? */
#include "simulink.c"      /* MEX-file interface mechanism */
#else
#include "cg_sfun.h"       /* Code generation registration function */
#endif
